//==============================================================================
//		Realtime Integer Library of Equations (RILE)
//		A R I T H M E T I C

//	Copyright (C) 2003  Ronald Alan Riley, PhD.
//	USC Information Sciences Institute

//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//$$  This program is free software; you can redistribute it and/or modify
//$$  it under the terms of the GNU General Public License as published by
//$$  the Free Software Foundation; either version 2 of the License, or
//$$  (at your option) any later version.

//$$  This program is distributed in the hope that it will be useful,
//$$  but WITHOUT ANY WARRANTY; without even the implied warranty of
//$$  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//$$  GNU General Public License for more details.

//$$  You should have received a copy of the GNU General Public License
//$$  along with this program; if not, write to the Free Software
//$$  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

//	This development was begun under DARPA's Power Aware Computing / Communication.
//	This file contains code for fast signal processing in integer math.
//	The purpose is to manage accuracy vs. processing time / battery energy.
//==============================================================================

//	Avoid branches in conditional asignments by using
//		(i>>31)     = -1 for i<0,  0 for i>=0
//		j & (i>>31) =  j for i<0,  0 for i>=0
//		j & ~(i>>31)=  0 for i<0,  j for i>=0

#ifndef RILE_A
#define RILE_A

#define Pi  3.1415926535897932384626433832795028841971693993751
#define Pi2 6.2831853071795864769252867665590057683843387987502
#define Pi_2_29 1686629713	/* Pi * 2^29 */

// Integer division to nearest integer. (U:Unknown, N:Negative, P:Positive)
#define NIntPP(n,d) (((d>>1)+n)/d)
#define NIntUP(n,d) (((d>>1)+n)/d+(((d>>1)+n-1)>>31))
#define NIntUN(n,d) -NIntUP(n,(-d))
#define NInt(n,d)   ((d<0) ? NIntUN(n,d) : NIntUP(n,d))

// Divide by powers of 2 by bit shifting, to nearest int.
#define ShRP1(n,s)	((1+(n>>s))>>1)
#define ShR1(n,s)	((1+((n+(n>>31))>>s))>>1)
#define ShRP(n,s)	ShRP1(n,(s-1))
#define ShR(n,s)	ShRP(n,(s-1))
#define ShRP0(n)	((1+n)>>1)
#define ShR0(n)		((1+n+(n>>31))>>1)

// Multiply by a fraction f = x * (2^31)  where |x| < 1.0f
// The other factor, i, can be an integer or fraction.
//#define iFracMultPP(f,i) (((f>>16)*(i>>16)<<1)+(((f>>16)*(i&0xFFFF))>>15)+(((i>>16)*(f&0xFFFF))>>15))
#define iFracMultPP(f,i) (((f>>16)*(i>>16)<<1) + ShRP1(((f>>16)*((i>>2)&0x3FFF)+\
	(i>>16)*((f>>2)&0x3FFF)+ ((((i>>1)&0x7FFF)*((f>>1)&0x7FFF))>>16)), 12))
#define iFracMultUP(f,i) ((f<0) ? -iFracMultPP((-f),i) : iFracMultPP(f,i))
#define iFracMultUN(f,i) ((f<0) ? iFracMultPP((-f),(-i)) : -iFracMultPP(f,(-i)))
#define iFracMult(f,i) ((i<0) ? iFracMultUN(f,i) : iFracMultUP(f,i))

// Multiply by fractions with power of 2 denominators. (Avoid overflow)
#define iFMultP0(i) 0						/* i * 0 */
#define iFMultN1(i) (-i)					/* i * (-1) */
#define iFMultP1(i) ( i)					/* i * (+1) */
#define iFMultN2_2(i) iFMultN1(i)			/* i * (-2/2) */
#define iFMultN1_2(i) ((-i)>>1)				/* i * (-1/2) */
#define iFMultP0_2(i) iFMultP0(i)			/* i * (+0/2) */
#define iFMultP1_2(i) (  i >>1)				/* i * (+1/2) */
#define iFMultP2_2(i) iFMultP1(i)			/* i * (+2/2) */
#define iFMultN4_4(i) iFMultN1(i)			/* i * (-4/4) */
#define iFMultN3_4(i) ((i>>2)-i)			/* i * (-3/4) */
#define iFMultN2_4(i) iFMultN1_2(i)			/* i * (-2/4) */
#define iFMultN1_4(i) ((-i)>>2)				/* i * (-1/4) */
#define iFMultP0_4(i) iFMultP0(i)			/* i * (+0/4) */
#define iFMultP1_4(i) (  i >>2)				/* i * (+1/4) */
#define iFMultP2_4(i) iFMultP1_2(i)			/* i * (+2/4) */
#define iFMultP3_4(i) (i-(i>>2))			/* i * (+3/4) */
#define iFMultP4_4(i) iFMultP1(i)			/* i * (+4/4) */
#define iFMultN8_8(i) iFMultN1(i)			/* i * (-8/8) */
#define iFMultN7_8(i) ((i>>3) - i)			/* i * (-7/8) */
#define iFMultN6_8(i) iFMultN3_4(i)			/* i * (-6/8) */
#define iFMultN5_8(i) (-(i>>3)-(i>>1))		/* i * (-5/8) */
#define iFMultN4_8(i) iFMultN1_2(i)			/* i * (-4/8) */
#define iFMultN3_8(i) ((i>>3)-(i>>1))		/* i * (-3/8) */
#define iFMultN2_8(i) iFMultN1_4(i)			/* i * (-2/8) */
#define iFMultN1_8(i) ((-i)>>3)				/* i * (-1/8) */
#define iFMultP0_8(i) iFMultP0(i)			/* i * (+0/8) */
#define iFMultP1_8(i) (  i >>3)				/* i * (+1/8) */
#define iFMultP2_8(i) iFMultP1_4(i)			/* i * (+2/8) */
#define iFMultP3_8(i) ((i>>1)-(i>>3))		/* i * (+3/8) */
#define iFMultP4_8(i) iFMultP1_2(i)			/* i * (+4/8) */
#define iFMultP5_8(i) ((i>>1)+(i>>3))		/* i * (+5/8) */
#define iFMultP6_8(i) iFMultP3_4(i)			/* i * (+6/8) */
#define iFMultP7_8(i) (i - (i>>3))			/* i * (+7/8) */
#define iFMultP8_8(i) iFMultP1(i)			/* i * (+8/8) */
#define iFMultN16_16(i) iFMultN1(i)			/* i * (-16/16) */
#define iFMultN15_16(i) ((i>>4)-i)			/* i * (-15/16) */
#define iFMultN14_16(i) iFMultN7_8(i)		/* i * (-14/16) */
#define iFMultN13_16(i) ((i>>3)+(i>>4)-i)	/* i * (-13/16) */
#define iFMultN12_16(i) iFMultN3_4(i)		/* i * (-12/16) */
#define iFMultN11_16(i) ((i>>2)+(i>>4)-i)	/* i * (-11/16) */
#define iFMultN10_16(i) iFMultN5_8(i)		/* i * (-10/16) */
#define iFMultN9_16(i)  (-(i>>1)-(i>>4))	/* i * (- 9/16) */
#define iFMultN8_16(i)  iFMultN1_2(i)		/* i * (- 8/16) */
#define iFMultN7_16(i)  ((i>>4)-(i>>1))		/* i * (- 7/16) */
#define iFMultN6_16(i)  iFMultN3_8(i)		/* i * (- 6/16) */
#define iFMultN5_16(i)  (-(i>>2)-(i>>4))	/* i * (- 5/16) */
#define iFMultN4_16(i)  iFMultN1_4(i)		/* i * (- 4/16) */
#define iFMultN3_16(i)  ((i>>4)-(i>>3))		/* i * (- 3/16) */
#define iFMultN2_16(i)  iFMultN1_8(i)		/* i * (- 2/16) */
#define iFMultN1_16(i)  ((-i)>>4)			/* i * (- 1/16) */
#define iFMultP0_16(i)  iFMultP0(i)			/* i * (+ 0/16) */
#define iFMultP1_16(i)  (i>>4)				/* i * (+ 1/16) */
#define iFMultP2_16(i)  iFMultP1_8(i)		/* i * (+ 2/16) */
#define iFMultP3_16(i)  ((i>>2)-(i>>4))		/* i * (+ 3/16) */
#define iFMultP4_16(i)  iFMultP1_4(i)		/* i * (+ 4/16) */
#define iFMultP5_16(i)  ((i>>2)+(i>>4))		/* i * (+ 5/16) */
#define iFMultP6_16(i)  iFMultP3_8(i)		/* i * (+ 6/16) */
#define iFMultP7_16(i)  ((i>>1)-(i>>4))		/* i * (+ 7/16) */
#define iFMultP8_16(i)  iFMultP1_2(i)		/* i * (+ 8/16) */
#define iFMultP9_16(i)  ((i>>1)+(i>>4))		/* i * (+ 9/16) */
#define iFMultP10_16(i) iFMultP5_8(i)		/* i * (+10/16) */
#define iFMultP11_16(i) (i-(i>>2)-(i>>4))	/* i * (+11/16) */
#define iFMultP12_16(i) iFMultP3_4(i)		/* i * (+12/16) */
#define iFMultP13_16(i) (i-(i>>3)-(i>>4))	/* i * (+13/16) */
#define iFMultP14_16(i) iFMultP7_8(i)		/* i * (+14/16) */
#define iFMultP15_16(i) (i-(i>>4))			/* i * (+15/16) */
#define iFMultP16_16(i) iFMultP1(i)			/* i * (+16/16) */

//	Absolutue Value
#define iAbs(i,t) ((i^(t=(i>>31)))-t)

// Swap two values
#define SWAP_t(a,b,t) {t=a; a=b; b=t;}
#define SWAP(a,b) {a ^= (b ^= (a ^= b));}

//	Binary serch for the number of bits required for positive input value
//	Return = Floor of Log2(i)
#define NumBits_2(i) (i>>1)
#define NumBits_4(i) ((i>>2) ? 2+(i>>3) : NumBits_2(i))
#define NumBits_8(i) ((i>>4) ? ((i>>6) ? 6+(i>>7) : 4+(i>>5)) : NumBits_4(i))
#define NumBits16(i) ((i>>8) ? \
((i>>12) ? ((i>>14) ? 14+(i>>15) :12+(i>>13)) : \
		   ((i>>10) ? 10+(i>>11) : 8+(i>>9))) : NumBits_8(i))
#define NumBits32(i) ((i>>16) ? \
((i>>24) ?	((i>>28) ? ((i>>30) ? 30+(i>>31) :28+(i>>29)) : \
					   ((i>>26) ? 26+(i>>27) :24+(i>>25))) : \
			((i>>20) ? ((i>>22) ? 22+(i>>23) :20+(i>>21)) : \
					   ((i>>18) ? 18+(i>>19) :16+(i>>17)))) : NumBits16(i))
#define NumBits(i) NumBits32(i)	/* Default to 32-bits (unsigned) */

#endif // RILE_A
