//==============================================================================
//		Realtime Integer Library of Equations (RILE)
//		C O R D I C

//	Copyright (C) 2003  Ronald Alan Riley, PhD.
//	USC Information Sciences Institute

//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//$$  This program is free software; you can redistribute it and/or modify
//$$  it under the terms of the GNU General Public License as published by
//$$  the Free Software Foundation; either version 2 of the License, or
//$$  (at your option) any later version.

//$$  This program is distributed in the hope that it will be useful,
//$$  but WITHOUT ANY WARRANTY; without even the implied warranty of
//$$  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//$$  GNU General Public License for more details.

//$$  You should have received a copy of the GNU General Public License
//$$  along with this program; if not, write to the Free Software
//$$  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

//	This development was begun under DARPA's Power Aware Computing / Communication.
//	This file contains code for fast signal processing in integer math.
//	The purpose is to manage accuracy vs. processing time / battery energy.
//==============================================================================
#ifndef RILE_C
#define RILE_C

//	Include Realtime Integer Library of Equations (RILE) for Arithmetic
#include "RILE_A.h"

//------------------------------------------------------------------------------
//	CORDIC coordinate rotation algorithm
//	Assume:	32-bit signed registers for calculations
//			16-bit signed input values for coordinates

// The CORDIC algorithm inflates the real/imaginary by
//		sqrt{2*(5/4)*(17/16)* ... *(1+2^14)/2^14}
#define CORDICSclF 1.646760257892

#define Two13mCORDIC     13490	/* ~ 1<<13 * CORDICScl */
#define Two12mCORDIC      6745	/* ~ 1<<12 * CORDICScl */
#define Two11mCORDIC      3373	/* ~ 1<<11 * CORDICScl */
#define Two10mCORDIC      1686	/* ~ 1<<10 * CORDICScl */
#define Two8mCORDIC        422	/* ~ 1<<8 * CORDICScl */

#define Two30dCORDIC 652032874	/* ~ 1<<30 / CORDICScl */
#define Two14dCORDIC      9949	/* ~ 1<<14 / CORDICScl */
#define Two12dCORDIC      2487	/* ~ 1<<12 / CORDICScl */
#define Two10dCORDIC       622	/* ~ 1<<10 / CORDICScl */
#define Two9dCORDIC        311	/* ~ 1<<9  / CORDICScl */

// Values > CORDIC_Max may overflow in CORDIC algorithms.
#define CORDIC_Max 922113733	/* ~2^31/sqrt(2)/CORDICScale */

//	Compute magnitude of complex number input as two 32-bit signed integer registers
//		R = sqrt(R*R + I*I) * CORDICSclF			I ~ 0
//	WARNING:  Overflow likely if input |R|,|I| > 1<<29
//	This function is optimized for speed to find complex mag, phase is not estimated.
#define CompMagStep(R, I, dR, n) \
dR=ShRP1(R,n);  R += ShRP1(I,n);  if (dR>>n) { if (I<dR) I=dR-I;  else I-=dR;
//dR=R>>(n+1);  R += I>>(n+1);  if (dR>>n) { if (I<dR) I=dR-I;  else I-=dR;
#define CompMag(R, I, dR) \
R = iAbs(R,dR); I = iAbs(I,dR); \
dR = R;          R += I;        if (I<dR) I=dR-I;  else I-=dR; \
dR = (1+R)>>1;   R += (1+I)>>1; if (I<dR) I=dR-I;  else I-=dR; \
CompMagStep(R, I, dR,  1) \
CompMagStep(R, I, dR,  2) \
CompMagStep(R, I, dR,  3) \
CompMagStep(R, I, dR,  4) \
CompMagStep(R, I, dR,  5) \
CompMagStep(R, I, dR,  6) \
CompMagStep(R, I, dR,  7) \
CompMagStep(R, I, dR,  8) \
CompMagStep(R, I, dR,  9) \
CompMagStep(R, I, dR, 10) \
CompMagStep(R, I, dR, 11) \
CompMagStep(R, I, dR, 12) \
CompMagStep(R, I, dR, 13) \
R += ShRP1(I,14); } } } } } } } } } } } } }

// CORDIC angles = (int) ((Atan(2^-n) in radians [-pi, pi))  * (1<<31 / pi))
#define CORDIC2Deg 8.381903171539E-08	/* 90.0/(1<<30) */
#define CORDIC2Rad 1.462918079267E-09	/* Pi/(1<<31) */
#define ATn2_99 1073741824	/* Atan(inf)  = 90.0 */
#define ATn2_00  536870912	/* Atan(2^0)  = 45.0 */
#define ATn2_01  316933406	/* Atan(2^-1) ~ 26.6 */
#define ATn2_02  167458907	/* Atan(2^-2) ~ 14.0 */
#define ATn2_03   85004756	/* Atan(2^-3) ~  7.1 */
#define ATn2_04   42667331	/* Atan(2^-4) ~  3.6 */
#define ATn2_05   21354465	/* Atan(2^-5) ~  1.8 */
#define ATn2_06   10679838	/* Atan(2^-6) ~  0.9 */
#define ATn2_07    5340245	/* Atan(2^-7) ~  0.4 */
#define ATn2_08    2670163	/* Atan(2^-8) ~  0.2 */
#define ATn2_09    1335087	/* Atan(2^-9) ~  0.1 */
#define ATn2_10     667544
#define ATn2_11     333772
#define ATn2_12     166886
#define ATn2_13      83443
#define ATn2_14      41722
#define ATn2_15      20861
#define ATn2_16      10430
#define ATn2_17       5215
#define ATn2_18       2608
#define ATn2_19       1304
#define ATn2_20        652
#define ATn2_21        326
#define ATn2_22        163
#define ATn2_23         81
#define ATn2_24         41
#define ATn2_25         20
#define ATn2_26         10
#define ATn2_27          5
#define ATn2_28          3
#define ATn2_29          1

//	Compute phase angle of integer complex number by rotation with CORDIC.
//	Magnitude is saved in real component, imaginary is driven toward zero.
//		R ~ sqrt(R*R + I*I) * CORDICSclF			I ~ 0
//	WARNING:  Overflow likely if input |R|,|I| > 1<<29
//	Output angle A = (int) (angle in radians [-pi, pi))  * (1<<31 / pi)
//	This function is optimized for speed to find atan as phase angle,
//	complex mag is only approximate

//	Angle can be improved by small angle correction based on residual value of I
//		iATan2(R, I, A, t);
//		A += I;	// ~ (I/R) ((1<<31)/Pi)
//	where R has been multiplied to the range (1.6*2^29,2^31) ~ 2^30, for I != 0

#define iATan2Step(R, I, A, dR, n, C, x) \
if(I){dR=ShRP1(R,n); if(I<-(dR>>1)){R+=ShRP1(-I,n); I+=dR; A-=C;} \
                else if(I>  dR>>1) {R+=ShRP1( I,n); I-=dR; A+=C;} x}
#define iATan2__3(R, I, A, dR, x) \
A=0; if(R<0){ R = -R;  I = -I;  A = (I<=0) ? 0x7FFFFFFF: 0x80000000;} \
if(I){dR=R;        if(I<0){R=-I;       I =dR; A-=ATn2_99;} else{R =I;       I=-dR; A+=ATn2_99;} \
if (dR<R) dR=R;  dR=NumBits(dR); \
if(dR>=29) {dR-=29; R=ShRP1(R,dR); I=ShR1(I,dR);} else if(dR<28) {dR=28-dR; R<<=dR; I<<=dR;} \
if(I){dR=R;        if(I<-(R>>1)) {R-=I;       I+=dR; A-=ATn2_00;} \
              else if(I>(R>>1))  {R+=I;       I-=dR; A+=ATn2_00;} \
if(I){dR=(1+R)>>1; if(I<-(dR>>1)){R+=(1-I)>>1;I+=dR; A-=ATn2_01;} \
              else if(I>(dR>>1)) {R+=(1+I)>>1;I-=dR; A+=ATn2_01;}  x } } }

#define iATan2__4(R, I, A, dR, x) iATan2__3(R, I, A, dR, iATan2Step(R, I, A, dR, 1, ATn2_02, x))
#define iATan2__5(R, I, A, dR, x) iATan2__4(R, I, A, dR, iATan2Step(R, I, A, dR, 2, ATn2_03, x))
#define iATan2__6(R, I, A, dR, x) iATan2__5(R, I, A, dR, iATan2Step(R, I, A, dR, 3, ATn2_04, x))
#define iATan2__7(R, I, A, dR, x) iATan2__6(R, I, A, dR, iATan2Step(R, I, A, dR, 4, ATn2_05, x))
#define iATan2__8(R, I, A, dR, x) iATan2__7(R, I, A, dR, iATan2Step(R, I, A, dR, 5, ATn2_06, x))
#define iATan2__9(R, I, A, dR, x) iATan2__8(R, I, A, dR, iATan2Step(R, I, A, dR, 6, ATn2_07, x))
#define iATan2_10(R, I, A, dR, x) iATan2__9(R, I, A, dR, iATan2Step(R, I, A, dR, 7, ATn2_08, x))
#define iATan2_11(R, I, A, dR, x) iATan2_10(R, I, A, dR, iATan2Step(R, I, A, dR, 8, ATn2_09, x))
#define iATan2_12(R, I, A, dR, x) iATan2_11(R, I, A, dR, iATan2Step(R, I, A, dR, 9, ATn2_10, x))
#define iATan2_13(R, I, A, dR, x) iATan2_12(R, I, A, dR, iATan2Step(R, I, A, dR,10, ATn2_11, x))
#define iATan2_14(R, I, A, dR, x) iATan2_13(R, I, A, dR, iATan2Step(R, I, A, dR,11, ATn2_12, x))
#define iATan2_15(R, I, A, dR, x) iATan2_14(R, I, A, dR, iATan2Step(R, I, A, dR,12, ATn2_13, x))
#define iATan2_16(R, I, A, dR, x) iATan2_15(R, I, A, dR, iATan2Step(R, I, A, dR,13, ATn2_14, x))
#define iATan2_17(R, I, A, dR, x) iATan2_16(R, I, A, dR, iATan2Step(R, I, A, dR,14, ATn2_15, x))
#define iATan2_18(R, I, A, dR, x) iATan2_17(R, I, A, dR, iATan2Step(R, I, A, dR,15, ATn2_16, x))
#define iATan2_19(R, I, A, dR, x) iATan2_18(R, I, A, dR, iATan2Step(R, I, A, dR,16, ATn2_17, x))
#define iATan2_20(R, I, A, dR, x) iATan2_19(R, I, A, dR, iATan2Step(R, I, A, dR,17, ATn2_18, x))
#define iATan2_21(R, I, A, dR, x) iATan2_20(R, I, A, dR, iATan2Step(R, I, A, dR,18, ATn2_19, x))
#define iATan2_22(R, I, A, dR, x) iATan2_21(R, I, A, dR, iATan2Step(R, I, A, dR,19, ATn2_20, x))
#define iATan2_23(R, I, A, dR, x) iATan2_22(R, I, A, dR, iATan2Step(R, I, A, dR,20, ATn2_21, x))
#define iATan2_24(R, I, A, dR, x) iATan2_23(R, I, A, dR, iATan2Step(R, I, A, dR,21, ATn2_22, x))
#define iATan2_25(R, I, A, dR, x) iATan2_24(R, I, A, dR, iATan2Step(R, I, A, dR,22, ATn2_23, x))
#define iATan2_26(R, I, A, dR, x) iATan2_25(R, I, A, dR, iATan2Step(R, I, A, dR,23, ATn2_24, x))
#define iATan2_27(R, I, A, dR, x) iATan2_26(R, I, A, dR, iATan2Step(R, I, A, dR,24, ATn2_25, x))
#define iATan2(R, I, A, dR) iATan2_27(R, I, A, dR, {})

//	Atan2 Results can be improved by small angle correction based on residual value of I
//		A += I / ((R Pi)>>31) = I / (1 - (1 - ((R Pi) >> 31)))
//			~ I + I(1 -  ((R Pi) >> 31)))
//			~ 2I - (I R Pi) >> 31
#define iATan2_Correction(R, I, A) A += (I<<1) - ((I*(((R>>14)*12868)>>26))>>3);

//	Convert from euclidian to polar using CORDIC
//	R	register int	Input x (real) coordinate, and output magnitude R=sqrt(R*R+I*I)*CORDICSclF
//	I	register int	Input y (imag) coordinate, driven toward zero I ~ 0
//	A	register int	Output angle A = (int) (angle in radians [-pi, pi))  * (1<<31 / pi)
//	dR	register int	Temp
//	For best results, scale inputs to |R|,|I| ~ 1<<29
//	WARNING:  Overflow likely if input |R|,|I| > 1<<29
#define iPolarStep(R, I, A, dR, n, C, x) \
dR=R>>(n+1); if(I<0){R+=(-I)>>(n+1); I+=dR; A-=C;}else{R+=I>>(n+1); I-=dR; A+=C;} x
//dR=ShRP1(R,n); if(I<0){R+=ShRP1(-I,n); I+=dR; A-=C;}else{R+=ShRP1( I,n); I-=dR; A+=C;} x
#define iPolar__3(R, I, A, dR, x) \
A=0; if(R<0){ R = -R;  I = -I;  A = (I<=0) ? 0x7FFFFFFF: 0x80000000;} \
dR=R;    if(I<0){R=-I;       I =dR; A-=ATn2_99;}else{R =I;    I=-dR; A+=ATn2_99;} \
dR=R;    if(I<0){R-=I;       I+=dR; A-=ATn2_00;}else{R+=I;    I-=dR; A+=ATn2_00;} \
dR=R>>1; if(I<0){R+=(-I)>>1; I+=dR; A-=ATn2_01;}else{R+=I>>1; I-=dR; A+=ATn2_01;} x
//dR=(1+R)>>1;if(I<0){R+=(1-I)>>1;I+=dR; A-=ATn2_01;}else{R+=(1+I)>>1;I-=dR; A+=ATn2_01;} x

#define iPolar__4(R, I, A, dR, x) iPolar__3(R, I, A, dR, iPolarStep(R, I, A, dR, 1, ATn2_02, x))
#define iPolar__5(R, I, A, dR, x) iPolar__4(R, I, A, dR, iPolarStep(R, I, A, dR, 2, ATn2_03, x))
#define iPolar__6(R, I, A, dR, x) iPolar__5(R, I, A, dR, iPolarStep(R, I, A, dR, 3, ATn2_04, x))
#define iPolar__7(R, I, A, dR, x) iPolar__6(R, I, A, dR, iPolarStep(R, I, A, dR, 4, ATn2_05, x))
#define iPolar__8(R, I, A, dR, x) iPolar__7(R, I, A, dR, iPolarStep(R, I, A, dR, 5, ATn2_06, x))
#define iPolar__9(R, I, A, dR, x) iPolar__8(R, I, A, dR, iPolarStep(R, I, A, dR, 6, ATn2_07, x))
#define iPolar_10(R, I, A, dR, x) iPolar__9(R, I, A, dR, iPolarStep(R, I, A, dR, 7, ATn2_08, x))
#define iPolar_11(R, I, A, dR, x) iPolar_10(R, I, A, dR, iPolarStep(R, I, A, dR, 8, ATn2_09, x))
#define iPolar_12(R, I, A, dR, x) iPolar_11(R, I, A, dR, iPolarStep(R, I, A, dR, 9, ATn2_10, x))
#define iPolar_13(R, I, A, dR, x) iPolar_12(R, I, A, dR, iPolarStep(R, I, A, dR,10, ATn2_11, x))
#define iPolar_14(R, I, A, dR, x) iPolar_13(R, I, A, dR, iPolarStep(R, I, A, dR,11, ATn2_12, x))
#define iPolar_15(R, I, A, dR, x) iPolar_14(R, I, A, dR, iPolarStep(R, I, A, dR,12, ATn2_13, x))
#define iPolar_16(R, I, A, dR, x) iPolar_15(R, I, A, dR, iPolarStep(R, I, A, dR,13, ATn2_14, x))
#define iPolar_17(R, I, A, dR, x) iPolar_16(R, I, A, dR, iPolarStep(R, I, A, dR,14, ATn2_15, x))
#define iPolar_18(R, I, A, dR, x) iPolar_17(R, I, A, dR, iPolarStep(R, I, A, dR,15, ATn2_16, x))
#define iPolar_19(R, I, A, dR, x) iPolar_18(R, I, A, dR, iPolarStep(R, I, A, dR,16, ATn2_17, x))
#define iPolar_20(R, I, A, dR, x) iPolar_19(R, I, A, dR, iPolarStep(R, I, A, dR,17, ATn2_18, x))
#define iPolar_21(R, I, A, dR, x) iPolar_20(R, I, A, dR, iPolarStep(R, I, A, dR,18, ATn2_19, x))
#define iPolar_22(R, I, A, dR, x) iPolar_21(R, I, A, dR, iPolarStep(R, I, A, dR,19, ATn2_20, x))
#define iPolar_23(R, I, A, dR, x) iPolar_22(R, I, A, dR, iPolarStep(R, I, A, dR,20, ATn2_21, x))
#define iPolar_24(R, I, A, dR, x) iPolar_23(R, I, A, dR, iPolarStep(R, I, A, dR,21, ATn2_22, x))
#define iPolar_25(R, I, A, dR, x) iPolar_24(R, I, A, dR, iPolarStep(R, I, A, dR,22, ATn2_23, x))
#define iPolar_26(R, I, A, dR, x) iPolar_25(R, I, A, dR, iPolarStep(R, I, A, dR,23, ATn2_24, x))
#define iPolar_27(R, I, A, dR, x) iPolar_26(R, I, A, dR, iPolarStep(R, I, A, dR,24, ATn2_25, x))
#define iPolar_28(R, I, A, dR, x) iPolar_27(R, I, A, dR, iPolarStep(R, I, A, dR,25, ATn2_26, x))
#define iPolar_29(R, I, A, dR, x) iPolar_28(R, I, A, dR, iPolarStep(R, I, A, dR,26, ATn2_27, x))
#define iPolar_30(R, I, A, dR, x) iPolar_29(R, I, A, dR, iPolarStep(R, I, A, dR,27, ATn2_28, x))
#define iPolar(R, I, A, dR) iPolar_30(R, I, A, dR, {})

// Rotate phase of integer complex number by input angle using CORDIC
// A = (int) (angle in radians [-pi, pi))  * (1<<31 / pi)
#define RotatePhaseStep(R, I, A, dR, n, C) \
dR=ShR1(R,n); if(dR||(I>>n)){ if(A<0) {R+=ShR1(I,n); I-=dR; A+=C;} else{R-=ShR1(I,n); I+=dR; A-=C;}
#define RotatePhase(R, I, A, dR) \
dR=R; if(A<0) {R =I; I=-dR; A+=ATn2_99;} else {R=-I; I =dR; A-=ATn2_99;} \
dR=R; if(A<0) {R+=I; I-=dR; A+=ATn2_00;} else {R-=I; I+=dR; A-=ATn2_00;} \
RotatePhaseStep(R, I, A, dR, 0, ATn2_01) \
RotatePhaseStep(R, I, A, dR, 1, ATn2_02) \
RotatePhaseStep(R, I, A, dR, 2, ATn2_03) \
RotatePhaseStep(R, I, A, dR, 3, ATn2_04) \
RotatePhaseStep(R, I, A, dR, 4, ATn2_05) \
RotatePhaseStep(R, I, A, dR, 5, ATn2_06) \
RotatePhaseStep(R, I, A, dR, 6, ATn2_07) \
RotatePhaseStep(R, I, A, dR, 7, ATn2_08) \
RotatePhaseStep(R, I, A, dR, 8, ATn2_09) \
RotatePhaseStep(R, I, A, dR, 9, ATn2_10) \
RotatePhaseStep(R, I, A, dR,10, ATn2_11) \
RotatePhaseStep(R, I, A, dR,11, ATn2_12) \
RotatePhaseStep(R, I, A, dR,12, ATn2_13) \
RotatePhaseStep(R, I, A, dR,13, ATn2_14) \
RotatePhaseStep(R, I, A, dR,14, ATn2_15) \
RotatePhaseStep(R, I, A, dR,15, ATn2_16) \
RotatePhaseStep(R, I, A, dR,16, ATn2_17) \
RotatePhaseStep(R, I, A, dR,17, ATn2_18) \
RotatePhaseStep(R, I, A, dR,18, ATn2_19) \
RotatePhaseStep(R, I, A, dR,19, ATn2_20) \
RotatePhaseStep(R, I, A, dR,20, ATn2_21) \
RotatePhaseStep(R, I, A, dR,21, ATn2_22) \
RotatePhaseStep(R, I, A, dR,22, ATn2_23) \
RotatePhaseStep(R, I, A, dR,23, ATn2_24) \
RotatePhaseStep(R, I, A, dR,24, ATn2_25) \
RotatePhaseStep(R, I, A, dR,25, ATn2_26) \
RotatePhaseStep(R, I, A, dR,26, ATn2_27) \
RotatePhaseStep(R, I, A, dR,27, ATn2_28) \
RotatePhaseStep(R, I, A, dR,28, ATn2_29) }}}}} }}}}} }}}}} }}}}} }}}}} }}}}

// Rotate phase of integer complex number by input angle using CORDIC
// A = (int) (angle in radians [-pi, pi))  * (1<<31 / pi)
#define RotatePhase14Step(R, I, A, dR, n, C) \
dR=R>>n; if(A<0) {R+=I>>n; I-=dR; A+=C;} else{R-=I>>n; I+=dR; A-=C;}
#define RotatePhase14(R, I, A, dR) \
dR=R; if(A<0) {R =I; I=-dR; A+=ATn2_99;} else {R=-I; I =dR; A-=ATn2_99;} \
dR=R; if(A<0) {R+=I; I-=dR; A+=ATn2_00;} else {R-=I; I+=dR; A-=ATn2_00;} \
RotatePhase14Step(R, I, A, dR, 1, ATn2_01) \
RotatePhase14Step(R, I, A, dR, 2, ATn2_02) \
RotatePhase14Step(R, I, A, dR, 3, ATn2_03) \
RotatePhase14Step(R, I, A, dR, 4, ATn2_04) \
RotatePhase14Step(R, I, A, dR, 5, ATn2_05) \
RotatePhase14Step(R, I, A, dR, 6, ATn2_06) \
RotatePhase14Step(R, I, A, dR, 7, ATn2_07) \
RotatePhase14Step(R, I, A, dR, 8, ATn2_08) \
RotatePhase14Step(R, I, A, dR, 9, ATn2_09) \
RotatePhase14Step(R, I, A, dR,10, ATn2_10) \
RotatePhase14Step(R, I, A, dR,11, ATn2_11) \
RotatePhase14Step(R, I, A, dR,12, ATn2_12) \
RotatePhase14Step(R, I, A, dR,13, ATn2_13)

#endif // RILE_C
