//==============================================================================
//		Realtime Integer Library of Equations (RILE)
//		F O U R I E R   T R A N S F O R M S

//	Copyright (C) 2003  Ronald Alan Riley, PhD.
//	USC Information Sciences Institute

//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//$$  This program is free software; you can redistribute it and/or modify
//$$  it under the terms of the GNU General Public License as published by
//$$  the Free Software Foundation; either version 2 of the License, or
//$$  (at your option) any later version.

//$$  This program is distributed in the hope that it will be useful,
//$$  but WITHOUT ANY WARRANTY; without even the implied warranty of
//$$  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//$$  GNU General Public License for more details.

//$$  You should have received a copy of the GNU General Public License
//$$  along with this program; if not, write to the Free Software
//$$  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

//	This development was begun under DARPA's Power Aware Computing / Communication.
//	This file contains code for fast signal processing in integer math.
//	The purpose is to manage accuracy vs. processing time / battery energy.
//==============================================================================
#ifndef RILE_F
#define RILE_F

//	Include Realtime Integer Library of Equations (RILE) for CORDIC
#include "RILE_C.h"

//	Include Realtime Integer Library of Equations (RILE) for Statistics
#include "RILE_S.h"

#define MAX_FFT_LEN 64

// Compute complex FFT of 16-bit signed integer data in 32-bit array.
// WARNING: Input signal shoould be scaled to 16-bit signed range for best accuracy.
// Butterfly implemented with a 14-bit multiply for phase rotation.
// Assuming 32-bit signed integer registers.
// Right-shift on even iterations to prevent overflow.
// Return number of bits shifted ~log2(n)/2-1.
//int iFFT(int *data, int n);		// Forward FFT of Complex signal
int iFFTr(int *data, int n);	// Reverse FFT of Complex signal

// NRC RealFT algorithm implemented for integer data.
// Complex output is converted to real power spectrum in integer format.
//int iRealFT(int *data, int n);	// Forward FFT of Real signal
int iRealFTr(int *data, int n);	// Reverse FFT of Real signal

//	Compute power spectrum of real signal
int iPowSpec(int *data, int n, int MinF, int MaxF);

//	Compute non-zero coeffs of sinc operator.
void ConstSinc(int *Sinc, const int rad, const int NBit);

//	Apply sinc blur twice to a signal and downsample by 2.
void SincBlurDS2(int *Sig, int *St, const int NSamp, int *Sinc, const int rad, const int NBit);

//	Integer, +fractional period power spectral estimate based on square-wave coeffs
//	int *iPS			Output power spectra, (MaxT-MinT+1)*NFrac power values
//	int *Sig			Input signal
//	const int NSamp		Number of input samples ~ 256
//	const int MinT		Minimum period >= 3     (samples) ~ 6
//	const int MaxT		Maximum period <= NSamp (samples) ~ 32
//	const int NFrac		Number of fractional periods for each integer period. 1,2,3,4,...
void iPPS(int *iPS, int *Sig, const int NSamp, const int MinT, const int MaxT, const int NFrac);

//	Integer power spectral estimate based on square-wave coeffs
//	int *iPS			Output power spectra, NF power values
//	int *Sig			Input signal
//	const int NSamp		Number of input samples ~ 256
//	const int MinF		Minimum frequency > SampRat / NSamp
//	const int MaxF		Maximum frequency < SampRat / 2
//	const int NF		Number of frequency samples requested
//	const int SampRat	Sampling Rate ~ 1024 samples/sec.
void iPS(int *iPS, int *Sig, const int NSamp,
		 const int MinF, const int MaxF, const int NF, const int SampRat);

#endif // RILE_F
