//==============================================================================
//		Realtime Integer Library of Equations (RILE)
//		S I G N A L   S T A T I S T I C S

//	Copyright (C) 2003  Ronald Alan Riley, PhD.
//	USC Information Sciences Institute

//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//$$  This program is free software; you can redistribute it and/or modify
//$$  it under the terms of the GNU General Public License as published by
//$$  the Free Software Foundation; either version 2 of the License, or
//$$  (at your option) any later version.

//$$  This program is distributed in the hope that it will be useful,
//$$  but WITHOUT ANY WARRANTY; without even the implied warranty of
//$$  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//$$  GNU General Public License for more details.

//$$  You should have received a copy of the GNU General Public License
//$$  along with this program; if not, write to the Free Software
//$$  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

//	This development was begun under DARPA's Power Aware Computing / Communication.
//	This file contains code for fast signal processing in integer math.
//	The purpose is to manage accuracy vs. processing time / battery energy.
//==============================================================================
#ifndef RILE_S
#define RILE_S

//	Include Realtime Integer Library of Equations (RILE) for Arithmetic
#include "RILE_A.h"

//	Sum signal values
//	Sum		(data)			The returned Sum
//	pS		*(data)			Pointer variable to signal array
//	NS		const int		Number of samples to be averaged
//	i		register int	Counter variable for looping
#define Sig_Sum(Sum, pS, NS, i) \
for (Sum = *pS, i=NS; --i; ) Sum += *(++pS);  pS-=NS-1;

//	Find average value of integer signal.
//	Avg		int				The returned mean value
//	pS		register *int	Pointer variable to signal array
//	NS		const int		Number of samples to be averaged
//	i		register int	Counter variable for looping
#define Sig_MeanI(Avg, pS, NS, i) \
Sig_Sum(Avg, pS, NS, i) \
Avg = NIntUP(Avg, NS);

//	Find slope of integer signal.
//	Slp		int				The returned slope << NB ~ (log2(NS)-1)
//	NB		int				Returned number of bits slope is shifted
//	pS		register *int	Pointer variable to signal array
//	pS1		register *int	Temp pointer variable
//	NS		const int		Number of samples to be averaged
//	i		register int	Counter variable for looping
//	Note: Sum(i^2)=n*(n+1)*(2n+1)/6
#define Sig_SlopeI(Slp, NB, pS, pS1, NS, i) \
for (pS1=pS+NS, Slp=0, i=1+(NB=NS>>1); --i; ) Slp += i*(*(--pS1) - *(pS++));  pS-=NB; \
Slp = (Slp*3)/((NB+NB+1)*(NB+1));  Slp <<= (NB=NumBits(NB));  Slp /= NS>>1;

//	Find average value of float signal.
//	Avg		float			The returned mean value
//	pS		*float			Pointer variable to signal array
//	NS		const int		Number of samples to be averaged
//	i		register int	Counter variable for looping
#define Sig_MeanF(Avg, pS, NS, i) \
Sig_Sum(Avg, pS, NS, i) \
Avg /= NS;

//	Find maximum difference from average
//	Max		(data)			The returned max deviation value
//	pS		*(data)			Pointer variable to signal array
//	NS		const int		Number of samples to be checked
//	i		register int	Counter variable for looping
#define Sig_MaxDev(Max, pS, NS, i) \
for (Max=(*pS<0)?-*pS:*pS, i=NS; --i; ) \
	if      (Max < *(++pS)) Max = *pS; \
	else if (Max <   -*pS)  Max = - *pS; \
pS-=NS-1;

//	Find min and max of signal
//	Min		(data)			The returned min value
//	Max		(data)			The returned max value
//	pS		*(data)			Pointer variable to signal array
//	NS		const int		Number of samples to be checked
//	i		register int	Counter variable for looping
#define Sig_Range(Min, Max, pS, NS, i) \
for (Min=Max= *pS, i=NS; --i; ) \
{	if (Max < *(++pS)) Max = *pS; \
	if (Min >    *pS)  Min = *pS; \
}   pS-=NS-1;

//	Add <Val> to sorted list, lengthening list by one
//	Val		register int	Value of element added to sorted list
//	Dat		*int			Data array
//	N		int				Number of sorted elements, should be incremented due to adding Val
//	pD		register *int	pointer to element in Data array
#define Sort_Add(Val, Dat, N, pD) \
{ for (pD=Dat+N; pD-- > Dat; ) if (*pD < Val) pD[1] = *pD;  else { pD[1]=Val; break;} \
if (pD < Dat) pD[1]=Val; }

//	Delete first occurance of <Val> from sorted list, reducing list by one
//	Val		register int	Value of element deleteted from sorted list
//	Dat		*int			Data array
//	N		int				Number of sorted elements, should be decremented due to removing Val
//	pD		register *int	pointer to element in Data array
#define Sort_Del(Val, Dat, N, pD) \
{ pD=Dat; while (*pD>Val) pD++; while (pD<Dat+(N-1)) *(pD++)=pD[1]; }

//	Bubble Sort
//	Dat		const *int		Data array
//	N		const int		Number of elements to be sorted
//	i		int				Number of elements currently sorted
//	pD		register *int	pointer to element in Data array
//	Val		register int	Value of element added to sorted list
#define BSort(Dat, N, i, pD, Val) \
{ for(i=1; i<N; i++) { Val=Dat[i]; Sort_Add(Val, Dat, i, pD) } }

#endif // RILE_S
