//==============================================================================
//		Realtime Integer Library of Equations (RILE)
//		F O U R I E R  T R A N S F O R M   Real Forward, 7-bit signal

//	Copyright (C) 2003  Ronald Alan Riley, PhD.
//	USC Information Sciences Institute

//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//$$  This program is free software; you can redistribute it and/or modify
//$$  it under the terms of the GNU General Public License as published by
//$$  the Free Software Foundation; either version 2 of the License, or
//$$  (at your option) any later version.

//$$  This program is distributed in the hope that it will be useful,
//$$  but WITHOUT ANY WARRANTY; without even the implied warranty of
//$$  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//$$  GNU General Public License for more details.

//$$  You should have received a copy of the GNU General Public License
//$$  along with this program; if not, write to the Free Software
//$$  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

//		gcc -c -O3 RILE_F.c

//	This development was begun under DARPA's Power Aware Computing / Communication.
//	This file contains code for fast signal processing in integer math.
//	The purpose is to manage accuracy vs. processing time / battery energy.
//==============================================================================
#include "RILE_F.h"

//	H1[n] =   (H[n] + H[N/2-n]*) / 2
//	H2[n] = i(-H[n] + H[N/2-n]*) / 2
//	F[n]   = H1[n]  + H2[n]  exp( 2 Pi i n / N)		Lower  freqs n<N/2
//	F[N-n] = H1[n]* - H2[n]* exp(-2 Pi i n / N)		Higher freqs n>N/2
#define iRealFT_Step(M, wr, wi) \
pS2 = (pS1 = data) + n-M;  pS1 +=M;\
reg_r = ( pS1[1] +  pS2[1]) >> 1; \
reg_i = (*pS2    - *pS1   ) >> 1; \
h2r = wr(reg_r) - wi(reg_i); \
h2i = wr(reg_i) + wi(reg_r); \
reg_r = (*pS1 + *pS2) >> 1; \
reg_i = (pS1[1] - pS2[1]) >> 1; \
*pS1   = h2r + reg_r; \
pS1[1] = h2i + reg_i; \
*pS2   = reg_r - h2r; \
pS2[1] = h2i - reg_i;

// RealFT algorithm implemented for integer data.
// Complex output is converted to real power spectrum in integer format.
char iRealFT(signed char *data, short n)
{	signed char h2r,h2i;
	char NShift;
	register signed char reg_r;
	register signed char reg_i;
	register signed char *pS1;
	register signed char *pS2;

//	Compute complex FFT in integer math.
//	Assume data has been scaled to 16-bit signed integer.
	NShift = iFFT(data, n);

//	Force signal to be 7 bit
	for (NShift++, pS1=data+n; pS1-- != data; )
		*pS1 >>= 1;

//	Process lowest-frequency complex components seperately.
	pS1 = data;
	reg_r = *pS1;
	reg_i =  pS1[1];
	*pS1   = reg_r + reg_i;		// Freq = 0
	pS1[1] = reg_r - reg_i;		// Freq = N/2
	if (n==4) return NShift;

//	Real integer FFT steps
#if MAX_FFT_LEN >=     8
//	Loop   2
	iRealFT_Step((n>> 2), iFMultP11_16, iFMultP11_16)
	if (n==  8) return NShift;

#if MAX_FFT_LEN >=    16
//	Loop   3
	iRealFT_Step((n>> 3), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 3)*   3), iFMultP6_16, iFMultP15_16)
	if (n== 16) return NShift;

#if MAX_FFT_LEN >=    32
//	Loop   4
	iRealFT_Step((n>> 4), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 4)*   3), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 4)*   5), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 4)*   7), iFMultP3_16, iFMultP16_16)
	if (n== 32) return NShift;

#if MAX_FFT_LEN >=    64
//	Loop   5
	iRealFT_Step((n>> 5), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 5)*   3), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 5)*   5), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 5)*   7), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>> 5)*   9), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>> 5)*  11), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 5)*  13), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 5)*  15), iFMultP2_16, iFMultP16_16)
	if (n== 64) return NShift;

#if MAX_FFT_LEN >=   128
//	Loop   6
	iRealFT_Step((n>> 6), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 6)*   3), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 6)*   5), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>> 6)*   7), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 6)*   9), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 6)*  11), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 6)*  13), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>> 6)*  15), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 6)*  17), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 6)*  19), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>> 6)*  21), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 6)*  23), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 6)*  25), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 6)*  27), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>> 6)*  29), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 6)*  31), iFMultP1_16, iFMultP16_16)
	if (n==128) return NShift;

#if MAX_FFT_LEN >=   256
//	Loop   7
	iRealFT_Step((n>> 7), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>> 7)*   3), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 7)*   5), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 7)*   7), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 7)*   9), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>> 7)*  11), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>> 7)*  13), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 7)*  15), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 7)*  17), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 7)*  19), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 7)*  21), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 7)*  23), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>> 7)*  25), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 7)*  27), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>> 7)*  29), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>> 7)*  31), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 7)*  33), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 7)*  35), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>> 7)*  37), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>> 7)*  39), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 7)*  41), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>> 7)*  43), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 7)*  45), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 7)*  47), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 7)*  49), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 7)*  51), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 7)*  53), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>> 7)*  55), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>> 7)*  57), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 7)*  59), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 7)*  61), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 7)*  63), iFMultP0_16, iFMultP16_16)
	if (n==256) return NShift;

#if MAX_FFT_LEN >=   512
//	Loop   8
	iRealFT_Step((n>> 8), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>> 8)*   3), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 8)*   5), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 8)*   7), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 8)*   9), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 8)*  11), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 8)*  13), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 8)*  15), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 8)*  17), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 8)*  19), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>> 8)*  21), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>> 8)*  23), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>> 8)*  25), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 8)*  27), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 8)*  29), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 8)*  31), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 8)*  33), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 8)*  35), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>> 8)*  37), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 8)*  39), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 8)*  41), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 8)*  43), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 8)*  45), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 8)*  47), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 8)*  49), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 8)*  51), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 8)*  53), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>> 8)*  55), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>> 8)*  57), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>> 8)*  59), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 8)*  61), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 8)*  63), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>> 8)*  65), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>> 8)*  67), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 8)*  69), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 8)*  71), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>> 8)*  73), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>> 8)*  75), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>> 8)*  77), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 8)*  79), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 8)*  81), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 8)*  83), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 8)*  85), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 8)*  87), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 8)*  89), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 8)*  91), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 8)*  93), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)*  95), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)*  97), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)*  99), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)* 101), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)* 103), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)* 105), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)* 107), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>> 8)* 109), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 111), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 113), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 115), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 117), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 119), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 121), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 123), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 125), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 8)* 127), iFMultP0_16, iFMultP16_16)
	if (n==512) return NShift;

#if MAX_FFT_LEN >=  1024
//	Loop   9
	iRealFT_Step((n>> 9), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>> 9)*   3), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>> 9)*   5), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>> 9)*   7), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 9)*   9), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 9)*  11), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 9)*  13), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 9)*  15), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>> 9)*  17), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 9)*  19), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 9)*  21), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 9)*  23), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 9)*  25), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>> 9)*  27), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 9)*  29), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 9)*  31), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 9)*  33), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 9)*  35), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>> 9)*  37), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>> 9)*  39), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>> 9)*  41), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>> 9)*  43), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>> 9)*  45), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>> 9)*  47), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 9)*  49), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 9)*  51), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 9)*  53), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 9)*  55), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 9)*  57), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>> 9)*  59), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 9)*  61), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 9)*  63), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 9)*  65), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 9)*  67), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>> 9)*  69), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>> 9)*  71), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>> 9)*  73), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 9)*  75), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 9)*  77), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 9)*  79), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>> 9)*  81), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 9)*  83), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 9)*  85), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 9)*  87), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 9)*  89), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 9)*  91), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>> 9)*  93), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 9)*  95), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 9)*  97), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 9)*  99), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 9)* 101), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 9)* 103), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>> 9)* 105), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>> 9)* 107), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>> 9)* 109), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>> 9)* 111), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>> 9)* 113), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>> 9)* 115), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>> 9)* 117), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 9)* 119), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 9)* 121), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 9)* 123), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 9)* 125), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>> 9)* 127), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>> 9)* 129), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>> 9)* 131), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 133), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 135), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 137), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 139), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 141), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 143), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 145), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>> 9)* 147), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 149), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 151), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 153), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 155), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 157), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 159), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 161), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 163), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>> 9)* 165), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 167), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 169), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 171), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 173), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 175), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 177), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 179), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 181), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 183), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>> 9)* 185), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 187), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 189), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 191), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 193), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 195), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 197), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 199), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 201), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 203), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 205), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 207), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 209), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 211), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 213), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 215), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>> 9)* 217), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 219), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 221), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 223), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 225), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 227), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 229), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 231), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 233), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 235), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 237), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 239), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 241), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 243), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 245), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 247), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 249), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 251), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 253), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>> 9)* 255), iFMultP0_16, iFMultP16_16)
	if (n==1024) return NShift;

#if MAX_FFT_LEN >=  2048
//	Loop  10
	iRealFT_Step((n>>10), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>10)*   3), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>10)*   5), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>10)*   7), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>10)*   9), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>10)*  11), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  13), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  15), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  17), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  19), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  21), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  23), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  25), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  27), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  29), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>10)*  31), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  33), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  35), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  37), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  39), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  41), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  43), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  45), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  47), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  49), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  51), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>10)*  53), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  55), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  57), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  59), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  61), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  63), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  65), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  67), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  69), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  71), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>10)*  73), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  75), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  77), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  79), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  81), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  83), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  85), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  87), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  89), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  91), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>10)*  93), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)*  95), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)*  97), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)*  99), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 101), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 103), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 105), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 107), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 109), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 111), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 113), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>10)* 115), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 117), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 119), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 121), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 123), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 125), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 127), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 129), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 131), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 133), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 135), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>10)* 137), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 139), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 141), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 143), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 145), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 147), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 149), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 151), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 153), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 155), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 157), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 159), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>10)* 161), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 163), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 165), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 167), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 169), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 171), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 173), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 175), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 177), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 179), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 181), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>10)* 183), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 185), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 187), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 189), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 191), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 193), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 195), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 197), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 199), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 201), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 203), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 205), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 207), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>10)* 209), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 211), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 213), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 215), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 217), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 219), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 221), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 223), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 225), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 227), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 229), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 231), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 233), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>10)* 235), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 237), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 239), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 241), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 243), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 245), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 247), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 249), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 251), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 253), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 255), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 257), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 259), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 261), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>10)* 263), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 265), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 267), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 269), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 271), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 273), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 275), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 277), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 279), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 281), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 283), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 285), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 287), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 289), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 291), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>10)* 293), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 295), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 297), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 299), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 301), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 303), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 305), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 307), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 309), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 311), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 313), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 315), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 317), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 319), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 321), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 323), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 325), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 327), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>10)* 329), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 331), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 333), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 335), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 337), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 339), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 341), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 343), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 345), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 347), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 349), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 351), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 353), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 355), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 357), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 359), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 361), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 363), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 365), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 367), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 369), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>10)* 371), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 373), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 375), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 377), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 379), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 381), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 383), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 385), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 387), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 389), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 391), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 393), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 395), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 397), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 399), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 401), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 403), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 405), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 407), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 409), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 411), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 413), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 415), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 417), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 419), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 421), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 423), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 425), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 427), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 429), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>10)* 431), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 433), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 435), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 437), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 439), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 441), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 443), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 445), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 447), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 449), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 451), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 453), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 455), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 457), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 459), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 461), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 463), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 465), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 467), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 469), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 471), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 473), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 475), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 477), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 479), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 481), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 483), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 485), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 487), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 489), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 491), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 493), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 495), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 497), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 499), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 501), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 503), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 505), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 507), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 509), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>10)* 511), iFMultP0_16, iFMultP16_16)
	if (n==2048) return NShift;

#if MAX_FFT_LEN >=  4096
//	Loop  11
	iRealFT_Step((n>>11), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*   3), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*   5), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*   7), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*   9), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*  11), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*  13), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*  15), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*  17), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*  19), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>11)*  21), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  23), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  25), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  27), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  29), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  31), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  33), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  35), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  37), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  39), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  41), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  43), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  45), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  47), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  49), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  51), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  53), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  55), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  57), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  59), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  61), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>11)*  63), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  65), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  67), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  69), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  71), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  73), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  75), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  77), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  79), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  81), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  83), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  85), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  87), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  89), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  91), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  93), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  95), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  97), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)*  99), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)* 101), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>11)* 103), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 105), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 107), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 109), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 111), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 113), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 115), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 117), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 119), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 121), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 123), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 125), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 127), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 129), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 131), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 133), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 135), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 137), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 139), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 141), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 143), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>11)* 145), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 147), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 149), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 151), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 153), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 155), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 157), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 159), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 161), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 163), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 165), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 167), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 169), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 171), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 173), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 175), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 177), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 179), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 181), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 183), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 185), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>11)* 187), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 189), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 191), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 193), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 195), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 197), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 199), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 201), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 203), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 205), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 207), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 209), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 211), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 213), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 215), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 217), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 219), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 221), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 223), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 225), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 227), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>11)* 229), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 231), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 233), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 235), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 237), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 239), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 241), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 243), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 245), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 247), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 249), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 251), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 253), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 255), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 257), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 259), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 261), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 263), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 265), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 267), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 269), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 271), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>11)* 273), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 275), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 277), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 279), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 281), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 283), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 285), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 287), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 289), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 291), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 293), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 295), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 297), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 299), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 301), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 303), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 305), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 307), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 309), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 311), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 313), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 315), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 317), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>11)* 319), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 321), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 323), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 325), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 327), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 329), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 331), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 333), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 335), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 337), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 339), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 341), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 343), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 345), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 347), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 349), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 351), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 353), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 355), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 357), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 359), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 361), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 363), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 365), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>11)* 367), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 369), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 371), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 373), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 375), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 377), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 379), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 381), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 383), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 385), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 387), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 389), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 391), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 393), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 395), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 397), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 399), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 401), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 403), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 405), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 407), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 409), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 411), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 413), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>11)* 415), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 417), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 419), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 421), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 423), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 425), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 427), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 429), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 431), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 433), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 435), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 437), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 439), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 441), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 443), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 445), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 447), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 449), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 451), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 453), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 455), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 457), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 459), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 461), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 463), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 465), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>11)* 467), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 469), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 471), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 473), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 475), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 477), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 479), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 481), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 483), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 485), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 487), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 489), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 491), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 493), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 495), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 497), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 499), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 501), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 503), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 505), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 507), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 509), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 511), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 513), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 515), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 517), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 519), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 521), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>11)* 523), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 525), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 527), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 529), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 531), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 533), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 535), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 537), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 539), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 541), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 543), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 545), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 547), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 549), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 551), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 553), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 555), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 557), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 559), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 561), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 563), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 565), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 567), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 569), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 571), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 573), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 575), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 577), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 579), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 581), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 583), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>11)* 585), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 587), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 589), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 591), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 593), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 595), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 597), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 599), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 601), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 603), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 605), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 607), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 609), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 611), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 613), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 615), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 617), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 619), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 621), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 623), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 625), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 627), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 629), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 631), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 633), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 635), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 637), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 639), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 641), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 643), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 645), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 647), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 649), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 651), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 653), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>11)* 655), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 657), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 659), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 661), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 663), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 665), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 667), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 669), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 671), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 673), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 675), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 677), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 679), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 681), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 683), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 685), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 687), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 689), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 691), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 693), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 695), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 697), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 699), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 701), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 703), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 705), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 707), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 709), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 711), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 713), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 715), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 717), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 719), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 721), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 723), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 725), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 727), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 729), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 731), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 733), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 735), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 737), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 739), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>11)* 741), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 743), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 745), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 747), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 749), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 751), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 753), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 755), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 757), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 759), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 761), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 763), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 765), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 767), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 769), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 771), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 773), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 775), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 777), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 779), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 781), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 783), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 785), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 787), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 789), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 791), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 793), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 795), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 797), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 799), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 801), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 803), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 805), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 807), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 809), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 811), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 813), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 815), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 817), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 819), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 821), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 823), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 825), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 827), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 829), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 831), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 833), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 835), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 837), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 839), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 841), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 843), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 845), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 847), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 849), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 851), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 853), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 855), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 857), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 859), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>11)* 861), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 863), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 865), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 867), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 869), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 871), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 873), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 875), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 877), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 879), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 881), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 883), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 885), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 887), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 889), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 891), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 893), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 895), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 897), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 899), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 901), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 903), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 905), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 907), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 909), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 911), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 913), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 915), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 917), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 919), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 921), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 923), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 925), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 927), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 929), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 931), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 933), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 935), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 937), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 939), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 941), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 943), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 945), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 947), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 949), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 951), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 953), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 955), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 957), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 959), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 961), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 963), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 965), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 967), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 969), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 971), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 973), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 975), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 977), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 979), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 981), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 983), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 985), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 987), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 989), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 991), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 993), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 995), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 997), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)* 999), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1001), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1003), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1005), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1007), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1009), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1011), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1013), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1015), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1017), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1019), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1021), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>11)*1023), iFMultP0_16, iFMultP16_16)
	if (n==4096) return NShift;

#if MAX_FFT_LEN >=  8192
//	Loop  12
	iRealFT_Step((n>>12), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*   3), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*   5), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*   7), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*   9), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  11), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  13), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  15), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  17), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  19), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  21), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  23), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  25), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  27), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  29), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  31), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  33), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  35), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  37), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  39), iFMultP16_16, iFMultP0_16)
	iRealFT_Step(((n>>12)*  41), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  43), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  45), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  47), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  49), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  51), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  53), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  55), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  57), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  59), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  61), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  63), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  65), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  67), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  69), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  71), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  73), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  75), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  77), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  79), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  81), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  83), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  85), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  87), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  89), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  91), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  93), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  95), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  97), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)*  99), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 101), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 103), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 105), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 107), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 109), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 111), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 113), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 115), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 117), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 119), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 121), iFMultP16_16, iFMultP1_16)
	iRealFT_Step(((n>>12)* 123), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 125), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 127), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 129), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 131), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 133), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 135), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 137), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 139), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 141), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 143), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 145), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 147), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 149), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 151), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 153), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 155), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 157), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 159), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 161), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 163), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 165), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 167), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 169), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 171), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 173), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 175), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 177), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 179), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 181), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 183), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 185), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 187), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 189), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 191), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 193), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 195), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 197), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 199), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 201), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 203), iFMultP16_16, iFMultP2_16)
	iRealFT_Step(((n>>12)* 205), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 207), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 209), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 211), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 213), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 215), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 217), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 219), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 221), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 223), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 225), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 227), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 229), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 231), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 233), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 235), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 237), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 239), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 241), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 243), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 245), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 247), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 249), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 251), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 253), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 255), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 257), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 259), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 261), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 263), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 265), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 267), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 269), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 271), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 273), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 275), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 277), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 279), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 281), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 283), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 285), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 287), iFMultP16_16, iFMultP3_16)
	iRealFT_Step(((n>>12)* 289), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 291), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 293), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 295), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 297), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 299), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 301), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 303), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 305), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 307), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 309), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 311), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 313), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 315), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 317), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 319), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 321), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 323), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 325), iFMultP16_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 327), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 329), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 331), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 333), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 335), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 337), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 339), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 341), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 343), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 345), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 347), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 349), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 351), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 353), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 355), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 357), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 359), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 361), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 363), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 365), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 367), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 369), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 371), iFMultP15_16, iFMultP4_16)
	iRealFT_Step(((n>>12)* 373), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 375), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 377), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 379), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 381), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 383), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 385), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 387), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 389), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 391), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 393), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 395), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 397), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 399), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 401), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 403), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 405), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 407), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 409), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 411), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 413), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 415), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 417), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 419), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 421), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 423), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 425), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 427), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 429), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 431), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 433), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 435), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 437), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 439), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 441), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 443), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 445), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 447), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 449), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 451), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 453), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 455), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 457), iFMultP15_16, iFMultP5_16)
	iRealFT_Step(((n>>12)* 459), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 461), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 463), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 465), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 467), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 469), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 471), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 473), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 475), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 477), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 479), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 481), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 483), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 485), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 487), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 489), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 491), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 493), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 495), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 497), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 499), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 501), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 503), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 505), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 507), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 509), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 511), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 513), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 515), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 517), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 519), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 521), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 523), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 525), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 527), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 529), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 531), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 533), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 535), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 537), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 539), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 541), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 543), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 545), iFMultP15_16, iFMultP6_16)
	iRealFT_Step(((n>>12)* 547), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 549), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 551), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 553), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 555), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 557), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 559), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 561), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 563), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 565), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 567), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 569), iFMultP15_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 571), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 573), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 575), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 577), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 579), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 581), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 583), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 585), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 587), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 589), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 591), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 593), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 595), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 597), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 599), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 601), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 603), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 605), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 607), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 609), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 611), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 613), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 615), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 617), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 619), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 621), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 623), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 625), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 627), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 629), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 631), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 633), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 635), iFMultP14_16, iFMultP7_16)
	iRealFT_Step(((n>>12)* 637), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 639), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 641), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 643), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 645), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 647), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 649), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 651), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 653), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 655), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 657), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 659), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 661), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 663), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 665), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 667), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 669), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 671), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 673), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 675), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 677), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 679), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 681), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 683), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 685), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 687), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 689), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 691), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 693), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 695), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 697), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 699), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 701), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 703), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 705), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 707), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 709), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 711), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 713), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 715), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 717), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 719), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 721), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 723), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 725), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 727), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 729), iFMultP14_16, iFMultP8_16)
	iRealFT_Step(((n>>12)* 731), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 733), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 735), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 737), iFMultP14_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 739), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 741), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 743), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 745), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 747), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 749), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 751), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 753), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 755), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 757), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 759), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 761), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 763), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 765), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 767), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 769), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 771), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 773), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 775), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 777), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 779), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 781), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 783), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 785), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 787), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 789), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 791), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 793), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 795), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 797), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 799), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 801), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 803), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 805), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 807), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 809), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 811), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 813), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 815), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 817), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 819), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 821), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 823), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 825), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 827), iFMultP13_16, iFMultP9_16)
	iRealFT_Step(((n>>12)* 829), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 831), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 833), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 835), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 837), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 839), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 841), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 843), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 845), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 847), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 849), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 851), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 853), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 855), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 857), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 859), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 861), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 863), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 865), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 867), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 869), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 871), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 873), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 875), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 877), iFMultP13_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 879), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 881), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 883), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 885), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 887), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 889), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 891), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 893), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 895), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 897), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 899), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 901), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 903), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 905), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 907), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 909), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 911), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 913), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 915), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 917), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 919), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 921), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 923), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 925), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 927), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 929), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 931), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 933), iFMultP12_16, iFMultP10_16)
	iRealFT_Step(((n>>12)* 935), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 937), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 939), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 941), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 943), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 945), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 947), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 949), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 951), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 953), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 955), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 957), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 959), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 961), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 963), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 965), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 967), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 969), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 971), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 973), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 975), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 977), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 979), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 981), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 983), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 985), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 987), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 989), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 991), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 993), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 995), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 997), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)* 999), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1001), iFMultP12_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1003), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1005), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1007), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1009), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1011), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1013), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1015), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1017), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1019), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1021), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1023), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1025), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1027), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1029), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1031), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1033), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1035), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1037), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1039), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1041), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1043), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1045), iFMultP11_16, iFMultP11_16)
	iRealFT_Step(((n>>12)*1047), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1049), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1051), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1053), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1055), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1057), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1059), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1061), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1063), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1065), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1067), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1069), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1071), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1073), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1075), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1077), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1079), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1081), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1083), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1085), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1087), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1089), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1091), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1093), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1095), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1097), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1099), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1101), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1103), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1105), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1107), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1109), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1111), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1113), iFMultP11_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1115), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1117), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1119), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1121), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1123), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1125), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1127), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1129), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1131), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1133), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1135), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1137), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1139), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1141), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1143), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1145), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1147), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1149), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1151), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1153), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1155), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1157), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1159), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1161), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1163), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1165), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1167), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1169), iFMultP10_16, iFMultP12_16)
	iRealFT_Step(((n>>12)*1171), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1173), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1175), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1177), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1179), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1181), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1183), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1185), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1187), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1189), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1191), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1193), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1195), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1197), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1199), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1201), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1203), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1205), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1207), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1209), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1211), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1213), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1215), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1217), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1219), iFMultP10_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1221), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1223), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1225), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1227), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1229), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1231), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1233), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1235), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1237), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1239), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1241), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1243), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1245), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1247), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1249), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1251), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1253), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1255), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1257), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1259), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1261), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1263), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1265), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1267), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1269), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1271), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1273), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1275), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1277), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1279), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1281), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1283), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1285), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1287), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1289), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1291), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1293), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1295), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1297), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1299), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1301), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1303), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1305), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1307), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1309), iFMultP9_16, iFMultP13_16)
	iRealFT_Step(((n>>12)*1311), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1313), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1315), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1317), iFMultP9_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1319), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1321), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1323), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1325), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1327), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1329), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1331), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1333), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1335), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1337), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1339), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1341), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1343), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1345), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1347), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1349), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1351), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1353), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1355), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1357), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1359), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1361), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1363), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1365), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1367), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1369), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1371), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1373), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1375), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1377), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1379), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1381), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1383), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1385), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1387), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1389), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1391), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1393), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1395), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1397), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1399), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1401), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1403), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1405), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1407), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1409), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1411), iFMultP8_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1413), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1415), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1417), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1419), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1421), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1423), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1425), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1427), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1429), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1431), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1433), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1435), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1437), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1439), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1441), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1443), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1445), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1447), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1449), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1451), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1453), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1455), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1457), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1459), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1461), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1463), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1465), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1467), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1469), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1471), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1473), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1475), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1477), iFMultP7_16, iFMultP14_16)
	iRealFT_Step(((n>>12)*1479), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1481), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1483), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1485), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1487), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1489), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1491), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1493), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1495), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1497), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1499), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1501), iFMultP7_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1503), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1505), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1507), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1509), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1511), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1513), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1515), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1517), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1519), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1521), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1523), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1525), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1527), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1529), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1531), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1533), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1535), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1537), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1539), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1541), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1543), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1545), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1547), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1549), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1551), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1553), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1555), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1557), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1559), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1561), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1563), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1565), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1567), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1569), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1571), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1573), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1575), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1577), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1579), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1581), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1583), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1585), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1587), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1589), iFMultP6_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1591), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1593), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1595), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1597), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1599), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1601), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1603), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1605), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1607), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1609), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1611), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1613), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1615), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1617), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1619), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1621), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1623), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1625), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1627), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1629), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1631), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1633), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1635), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1637), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1639), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1641), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1643), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1645), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1647), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1649), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1651), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1653), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1655), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1657), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1659), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1661), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1663), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1665), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1667), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1669), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1671), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1673), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1675), iFMultP5_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1677), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1679), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1681), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1683), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1685), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1687), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1689), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1691), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1693), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1695), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1697), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1699), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1701), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1703), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1705), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1707), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1709), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1711), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1713), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1715), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1717), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1719), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1721), iFMultP4_16, iFMultP15_16)
	iRealFT_Step(((n>>12)*1723), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1725), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1727), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1729), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1731), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1733), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1735), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1737), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1739), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1741), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1743), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1745), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1747), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1749), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1751), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1753), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1755), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1757), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1759), iFMultP4_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1761), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1763), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1765), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1767), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1769), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1771), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1773), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1775), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1777), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1779), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1781), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1783), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1785), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1787), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1789), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1791), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1793), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1795), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1797), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1799), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1801), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1803), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1805), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1807), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1809), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1811), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1813), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1815), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1817), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1819), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1821), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1823), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1825), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1827), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1829), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1831), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1833), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1835), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1837), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1839), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1841), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1843), iFMultP3_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1845), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1847), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1849), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1851), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1853), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1855), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1857), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1859), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1861), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1863), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1865), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1867), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1869), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1871), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1873), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1875), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1877), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1879), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1881), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1883), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1885), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1887), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1889), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1891), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1893), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1895), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1897), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1899), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1901), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1903), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1905), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1907), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1909), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1911), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1913), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1915), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1917), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1919), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1921), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1923), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1925), iFMultP2_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1927), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1929), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1931), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1933), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1935), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1937), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1939), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1941), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1943), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1945), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1947), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1949), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1951), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1953), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1955), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1957), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1959), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1961), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1963), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1965), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1967), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1969), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1971), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1973), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1975), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1977), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1979), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1981), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1983), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1985), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1987), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1989), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1991), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1993), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1995), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1997), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*1999), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2001), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2003), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2005), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2007), iFMultP1_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2009), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2011), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2013), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2015), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2017), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2019), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2021), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2023), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2025), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2027), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2029), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2031), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2033), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2035), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2037), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2039), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2041), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2043), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2045), iFMultP0_16, iFMultP16_16)
	iRealFT_Step(((n>>12)*2047), iFMultP0_16, iFMultP16_16)
	if (n==8192) return NShift;
#endif	// MAX_FFT_LEN >=  8192
#endif	// MAX_FFT_LEN >=  4096
#endif	// MAX_FFT_LEN >=  2048
#endif	// MAX_FFT_LEN >=  1024
#endif	// MAX_FFT_LEN >=   512
#endif	// MAX_FFT_LEN >=   256
#endif	// MAX_FFT_LEN >=   128
#endif	// MAX_FFT_LEN >=    64
#endif	// MAX_FFT_LEN >=    32
#endif	// MAX_FFT_LEN >=    16
#endif	// MAX_FFT_LEN >=     8
	return NShift;
}
#undef iRealFT_Step
